/* Helper functions for high degree Toom-Cook.

   Contributed to the GNU project by Marco Bodrato.

   THE FUNCTIONS IN THIS FILE ARE INTERNAL WITH A MUTABLE INTERFACE.
   IT IS ONLY SAFE TO REACH THEM THROUGH DOCUMENTED INTERFACES.
   IN FACT, IT IS ALMOST GUARANTEED THAT THEY WILL CHANGE OR
   DISAPPEAR IN A FUTURE GNU MP RELEASE.

   Copyright 2009 Free Software Foundation, Inc.

   This file is part of the GNU MP Library.

   The GNU MP Library is free software; you can redistribute it
   and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 3 of the License, or (at your option) any later version.

   The GNU MP Library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the GNU MP Library.
   If not, see http://www.gnu.org/licenses/.
 */

/* #define USE_MUL_1 1 */

#if HAVE_NATIVE_mpn_sublsh_n
#define DO_mpn_sublsh_n(dst,src,n,s,ws) mpn_sublsh_n(dst,src,n,s)
#else
static mp_limb_t
DO_mpn_sublsh_n(mp_ptr dst, mp_srcptr src, mp_size_t n, unsigned int s, mp_ptr ws)
{
#if USE_MUL_1
  return mpn_submul_1(dst,src,n,CNST_LIMB(1) <<(s));
#else
  mp_limb_t __cy = mpn_lshift(ws,src,n,s);
  return    __cy + mpn_sub_n(dst,dst,ws,n);
#endif
}
#endif

#if HAVE_NATIVE_mpn_addlsh_n
#define DO_mpn_addlsh_n(dst,src,n,s,ws) mpn_addlsh_n(dst,src,n,s)
#else
static mp_limb_t
DO_mpn_addlsh_n(mp_ptr dst, mp_srcptr src, mp_size_t n, unsigned int s, mp_ptr ws)
{
#if USE_MUL_1
  return mpn_addmul_1(dst,src,n,CNST_LIMB(1) <<(s));
#else
  mp_limb_t __cy = mpn_lshift(ws,src,n,s);
  return    __cy + mpn_add_n(dst,dst,ws,n);
#endif
}
#endif

#if HAVE_NATIVE_mpn_subrsh
#define DO_mpn_subrsh(dst,nd,src,ns,s,ws) mpn_subrsh(dst,nd,src,ns,s)
#else
/* This is not a correct definition, it assumes no carry */
#define DO_mpn_subrsh(dst,nd,src,ns,s,ws)				\
do {									\
  mp_limb_t __cy;							\
  MPN_DECR_U (dst, nd, src[0] >> s);					\
  __cy = DO_mpn_sublsh_n (dst, src + 1, ns - 1, GMP_NUMB_BITS - s, ws);	\
  MPN_DECR_U (dst + ns - 1, nd - ns + 1, __cy);				\
} while (0)
#endif


/* Stores |{ap,n}-{bp,n}| in {rp,n}, returns the sign. */
static int
abs_sub_n (mp_ptr rp, mp_srcptr ap, mp_srcptr bp, mp_size_t n)
{
  mp_limb_t  x, y;
  while (--n >= 0)
    {
      x = ap[n];
      y = bp[n];
      if (x != y)
	{
	  n++;
	  if(x > y)
	    {
	      mpn_sub_n (rp, ap, bp, n);
	      return 0;
	    }
	  else
	    {
	      mpn_sub_n (rp, bp, ap, n);
	      return 1;
	    }
	}
      rp[n] = 0;
    }
  return 0;
}

static void
toom_couple_handling(mp_ptr pp, mp_size_t n, mp_ptr np, int nsign, mp_size_t off, int ps, int ns)
{
  if (nsign) {
#ifdef HAVE_NATIVE_mpn_rsh1sub_n
    mpn_rsh1sub_n (np, pp, np, n);
#else
    mpn_sub_n (np, pp, np, n);
    mpn_rshift (np, np, n, 1);
#endif
  } else {
#ifdef HAVE_NATIVE_mpn_rsh1add_n
    mpn_rsh1add_n (np, pp, np, n);
#else
    mpn_add_n (np, pp, np, n);
    mpn_rshift (np, np, n, 1);
#endif
  }

#ifdef HAVE_NATIVE_mpn_rsh1sub_n
  if (ps == 1)
    mpn_rsh1sub_n (pp, pp, np, n);
  else
#endif
  {
    mpn_sub_n (pp, pp, np, n);
    if (ps > 0)
      mpn_rshift (pp, pp, n, ps);
  }
  if (ns > 0)
    mpn_rshift (np, np, n, ns);
  pp[n-1] += mpn_add_n(pp+off, pp+off, np, n>>1);
  ASSERT_NOCARRY( mpn_add_1(pp+n-1, np+off, n-off, pp[n-1]) );
}

static int
abs_sub_add_n (mp_ptr rm, mp_ptr rp, mp_srcptr rs, mp_size_t n) {
  int result = abs_sub_n (rm, rp, rs, n);
  ASSERT_NOCARRY(mpn_add_n (rp, rp, rs, n));
  return result;
}

static int
mpn_toom_ev_pm1(mp_ptr rp, mp_ptr rm,
		mp_srcptr ap, unsigned int q, mp_size_t n, mp_size_t t,
		mp_ptr ws)
{
  /* {ap,q*n+t} -> {rp,n+1} {rm,n+1} , with {ws, n+1}*/
  ASSERT( n >= t );
  ASSERT( q > 2 );
  if( (q & 1) == 0) {
    rp[n] = mpn_add(rp, ap+n*(q-2), n, ap+n*q, t);
    q--;
    ws[n] = mpn_add_n(ws, ap+n*(q-2), ap+n*q, n);
    q-=3;
    rp[n]+= mpn_add_n(rp, rp, ap+n*q, n);
  } else {
    ws[n] = mpn_add(ws, ap+n*(q-2), n, ap+n*q, t);
    q-=3;
    rp[n] = mpn_add_n(rp, ap+n*q, ap+n*(q+2), n);
  }
  while(q) {
    q--;
    ws[n] += mpn_add_n(ws, ws, ap+n*q, n);
    q--;
    rp[n] += mpn_add_n(rp, rp, ap+n*q, n);
  }
  return abs_sub_add_n (rm, rp, ws, n + 1);
}

static int
mpn_toom_ev_lsh(mp_ptr rp, mp_ptr rm,
		mp_srcptr ap, unsigned int q, mp_size_t n, mp_size_t t,
		unsigned int s, mp_ptr ws, mp_ptr scratch)
{
  /* {ap,q*n+t} -> {rp,n+1} {rm,n+1} , with {ws, n+1}*/
  ASSERT( n >= t );
  ASSERT( s != 0 ); /* or _ev_pm1 should be used */
  ASSERT( q > 2 );
  ASSERT( s*q < GMP_NUMB_BITS );
  if( (q & 1) == 0) {
    rp[t] = mpn_lshift(rp, ap+n*q, t, s*q);
    MPN_ZERO (rp + t + 1, n - t);
    q--;
    ws[n] = mpn_lshift(ws, ap+n*q, n, s*q);
    q--;
    rp[n] += DO_mpn_addlsh_n(rp, ap+n*q, n, s*q, scratch);
  } else {
    ws[t] = mpn_lshift(ws, ap+n*q, t, s*q);
    MPN_ZERO(ws + t + 1, n - t);
    q--;
    rp[n] = mpn_lshift(rp, ap+n*q, n, s*q);
  }
  do {
    q--;
    ws[n] += DO_mpn_addlsh_n(ws, ap+n*q, n, s*q, scratch);
    q--;
    if(q != 0)
      rp[n] += DO_mpn_addlsh_n(rp, ap+n*q, n, s*q, scratch);
    else {
      rp[n] += mpn_add_n(rp, rp, ap, n);
      break;
    }
  } while (1);
  return abs_sub_add_n (rm, rp, ws, n + 1);
}

static int
mpn_toom_ev_rsh(mp_ptr rp, mp_ptr rm,
		mp_srcptr ap, unsigned int q, mp_size_t n, mp_size_t t,
		unsigned int s, mp_ptr ws, mp_ptr scratch)
{
  unsigned int i;
  /* {ap,q*n+t} -> {rp,n+1} {rm,n+1} , with {ws, n+1}*/
  ASSERT( n >= t );
  ASSERT( s != 0 ); /* or _ev_pm1 should be used */
  ASSERT( q > 1 );
  ASSERT( s*q < GMP_NUMB_BITS );
  rp[n] = mpn_lshift(rp, ap, n, s*q);
  //  dumpy(rp,n+1);
  ws[n] = mpn_lshift(ws, ap+n, n, s*(q-1));
  //  dumpy(ws,n);
  if( (q & 1) != 0) {
    ASSERT_NOCARRY(mpn_add(ws,ws,n+1,ap+n*q,t));
    //  dumpy(ws,n);
    rp[n] += DO_mpn_addlsh_n(rp, ap+n*(q-1), n, s, scratch);
    //  dumpy(rp,n+1);
  } else {
    ASSERT_NOCARRY(mpn_add(rp,rp,n+1,ap+n*q,t));
    //  dumpy(rp,n+1);
  }
  for(i=2; i<q-1; i++)
  {
    rp[n] += DO_mpn_addlsh_n(rp, ap+n*i, n, s*(q-i), scratch);
    //  dumpy(rp,n+1);
    i++;
    ws[n] += DO_mpn_addlsh_n(ws, ap+n*i, n, s*(q-i), scratch);
    //  dumpy(ws,n);
  };
  //  dumpy(rp,n+1);
  //  dumpy(ws,n);
  return abs_sub_add_n (rm, rp, ws, n + 1);
}

#ifdef WANT_TOOM_EV_MUL
static int
mpn_toom_ev_mul(mp_ptr rp, mp_ptr rm,
		mp_srcptr ap, unsigned int q, mp_size_t n, mp_size_t t,
		unsigned int s, mp_ptr ws)
{
  unsigned int m = s;
  ASSERT( n >= t );
  ASSERT( s > 2 );
  ASSERT( q > 2 );
  /* {ap,q*n+t} -> {rp,n+1} {rm,n+1} , with {ws, n+1}*/
  ap += n*q;
  {unsigned int i; for(i=q; --i; ) {m *= s;}; };
  ASSERT( (m*s-1)/(s-1) <= GMP_NUMB_MAX );
  if( (q & 1) == 0) {
    rp[t] = mpn_mul_1(rp, ap, t, m);
    MPN_ZERO (rp + t + 1, n - t);
    ap-=n;m/=s;
    ws[n] = mpn_mul_1(ws, ap, n, m);
    ap-=n;m/=s;
    rp[n] += mpn_addmul_1(rp, ap, n, m);
  } else {
    ws[t] = mpn_mul_1(ws, ap, t, m);
    MPN_ZERO(ws + t + 1, n - t);
    ap-=n;m/=s;
    rp[n] = mpn_mul_1(rp, ap, n, m);
  }
  do {
    ap-=n;m/=s;
    ws[n] += mpn_addmul_1(ws, ap, n, m);
    ap-=n;m/=s;
    if(m != 1)
      rp[n] += mpn_addmul_1(rp, ap, n, m);
    else {
      rp[n] += mpn_add_n(rp, rp, ap, n);
      break;
    }
  } while (1);
  return abs_sub_add_n (rm, rp, ws, n + 1);
}
#endif
