/* Implementation of the algorithm for Toom-Cook 4.5-way.

   Contributed to the GNU project by Marco Bodrato.

   THE FUNCTIONS IN THIS FILE ARE INTERNAL WITH A MUTABLE INTERFACE.
   IT IS ONLY SAFE TO REACH THEM THROUGH DOCUMENTED INTERFACES.
   IN FACT, IT IS ALMOST GUARANTEED THAT THEY WILL CHANGE OR
   DISAPPEAR IN A FUTURE GNU MP RELEASE.

   Copyright 2009 Free Software Foundation, Inc.

   This file is part of the GNU MP Library.

   The GNU MP Library is free software; you can redistribute it
   and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 3 of the License, or (at your option) any later version.

   The GNU MP Library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the GNU MP Library.
   If not, see http://www.gnu.org/licenses/.
 */

#include "gmp.h"
#include "gmp-impl.h"

#include "toom-tools.h"

/* definitions for divexact_by{9,15}, taken from toom_interpolate_7pts */

#define BINVERT_9 \
  ((((GMP_NUMB_MAX / 9) << (6 - GMP_NUMB_BITS % 6)) * 8 & GMP_NUMB_MAX) | 0x39)

#define BINVERT_15 \
  ((((GMP_NUMB_MAX >> (GMP_NUMB_BITS % 4)) / 15) * 14 * 16 & GMP_NUMB_MAX) + 15))

/* For the various mpn_divexact_byN here, fall back to using either
   mpn_bdiv_q_1_pi1 or mpn_divexact_1.  The former has less overhead and is
   many faster if it is native.  For now, since mpn_divexact_1 is native on
   several platforms where mpn_bdiv_q_1_pi1 does not yet exist, do not use
   mpn_bdiv_q_1_pi1 unconditionally.  FIXME.  */

#ifndef mpn_divexact_by9
#if HAVE_NATIVE_mpn_bdiv_q_1_pi1
#define mpn_divexact_by9(dst,src,size) mpn_bdiv_q_1_pi1(dst,src,size,9,BINVERT_9,0)
#else
#define mpn_divexact_by9(dst,src,size) mpn_divexact_1(dst,src,size,9)
#endif
#endif

#ifndef mpn_divexact_by15
#if HAVE_NATIVE_mpn_bdiv_q_1_pi1
#define mpn_divexact_by15(dst,src,size) mpn_bdiv_q_1_pi1(dst,src,size,15,BINVERT_15,0)
#else
#define mpn_divexact_by15(dst,src,size) mpn_divexact_1(dst,src,size,15)
#endif
#endif

/* Interpolation for Toom-4.5 (or Toom-4), using the evaluation
   points: infinity(4.5 only), 4, -4, 2, -2, 1, -1, 0. More precisely,
   we want to compute f(2^(GMP_NUMB_BITS * n)) for a polynomial f of
   degree 7 (or 6), given the 8 (rsp. 7) values:

     r1 = limit at infinity of f(x) / x^7,
     r2 = f(2),
     r3 = f(-2),
     r4 = f(1),
     r5 = f(-1),
     r6 = f(1/2),
     r7 = f(-1/2),
     r8 = f(0).

   All couples of the form f(n),f(-n) must be already mixed with
   toom_couple_handling(f(n),...,f(-n),...)

   The result is stored in {pp, spt + 7*n (or 6*n)}.
   At entry, r8 is stored at {pp, 2n},
   r5 is stored at {pp + 3n, 3n + 1}.

   The other values are 3n+1 limbs each (with most significant limbs small).

   All intermediate results are positive.
   Inputs are destroyed.
*/

static void
mpn_toom_interpolate_8pts (mp_ptr pp, mp_size_t n,
			   mp_ptr r3, mp_ptr r7,
			   mp_size_t spt, mp_ptr ws)
{
  mp_limb_t cy;

  mp_ptr    r5 = (pp + 3 * n);			/* 3n+1 */
  mp_ptr    r1 = (pp + 7 * n);			/* spt */

  /******************************* interpolation *****************************/

  r5[3*n] -= mpn_sub_n (r5+n, r5+n, pp, 2 * n);
  cy = mpn_sub_n (r5, r5, r1, spt);
  MPN_DECR_U (r5 + spt, 3 * n + 1 - spt, cy);

  DO_mpn_subrsh(r7, 3 * n + 1, r1, spt, 2, ws);
  r7[3*n] -= DO_mpn_sublsh_n (r7 + n, pp, 2*n, 6, ws);

  DO_mpn_subrsh(r3+n, 2 * n + 1, pp, 2 * n, 2, ws);
  cy = DO_mpn_sublsh_n (r3, r1, spt, 6, ws);
  MPN_DECR_U (r3 + spt, 3 * n + 1 - spt, cy);

  ASSERT_NOCARRY(mpn_add_n (r3, r3, r7, 3 * n + 1));
  ASSERT_NOCARRY(DO_mpn_sublsh_n (r3, r5, 3 * n + 1, 3, ws));
  mpn_divexact_by9 (r3, r3, 3 * n + 1);

  ASSERT_NOCARRY(mpn_sub_n (r5, r5, r3, 3 * n + 1));

  ASSERT_NOCARRY(mpn_sub_n (r7, r7, r3, 3 * n + 1));
  ASSERT_NOCARRY(DO_mpn_sublsh_n (r7, r5, 3 * n + 1, 2, ws));
  mpn_divexact_by15(r7, r7, 3 * n + 1);

  /* last interpolation step... */
  /* ... is mixed with recomposition */

  /***************************** recomposition *******************************/
  /*
    pp[] prior to operations:
     |_H r1|_L r1|____||_H r5|_M_r5|_L r5|_____|_H r8|_L r8|pp

    summation scheme for remaining operations:
     |____8|n___7|n___6|n___5|n___4|n___3|n___2|n____|n____|pp
     |_H r1|_L r1|____||_H_r5|_M r5|_L r5|_____|_H_r8|_L r8|pp
	  ||_H r3|_M r3|_L_r3|
				  ||_H_r7|_M_r7|_L_r7|
	  ||-H r7|-M r7|-L_r7|
  */

  mp_limb_t bw;
  
  cy = mpn_add_n (pp + 5*n, r5 + 2*n, r3, n + 1); /* Hr5+Lr3 */
  bw = mpn_sub_n (pp + 5*n, pp + 5*n, r7, n + 1); /* Hr5+Lr3-Lr7 */
  if (bw != cy) {
    if (UNLIKELY(bw > cy)) MPN_DECR_U (r3 + n + 1, 2*n, 1);
    else                   MPN_INCR_U (r3 + n + 1, 2*n, 1);
  }
  cy = mpn_sub_n (pp + 6*n + 1, r3 + n + 1, r7 + n + 1, n - 1);
  MPN_DECR_U (r3 + 2*n, n+1, cy);
  ASSERT_NOCARRY(mpn_sub_n (r3 + 2*n, r3 + 2*n, r7 + 2*n, n+1));
  cy = mpn_add_n (pp + 7*n, pp + 7*n, r3 + 2*n, n+1);
  ASSERT( (cy == 0) || (spt > n + 1) );
  if( spt > n + 1 )MPN_INCR_U (pp + 8*n+1, spt - (n + 1), cy);

  cy = mpn_add_n (pp + n, pp + n, r7, n); /* Hr8+Lr7 */
  cy = mpn_add_1 (pp + 2*n, r7+n, n, cy);
  MPN_INCR_U (r7 + 2*n, n + 1, cy);
  cy = mpn_add_n (pp + 3*n, r5, r7+ 2*n, n+1); /* Hr7+Lr5 */
  MPN_INCR_U (r5 + n + 1, 3*n - 1 + spt, cy);
}

/* Toom-4.5, the splitting 5x4 unbalanced version.
   Evaluate in: infinity, +4, -4, +2, -2, +1, -1, 0.

  <--s-><--n--><--n--><--n--><--n-->
   ____ ______ ______ ______ ______
  |_a4_|__a3__|__a2__|__a1__|__a0__|
          |b3_|__b2__|__b1__|__b0__|
          <-t-><--n--><--n--><--n-->

 WARNING! This code has some limitations, look at the first few ASSERTs!
*/
#define TOOM_54_MUL_N_REC(p, a, b, n, ws)		\
  do {	mpn_mul_n(p, a, b, n);				\
  } while (0)

#define TOOM_54_MUL_REC(p, a, na, b, nb, ws)		\
  do {	mpn_mul (p, a, na, b, nb);			\
  } while (0)

void
mpn_toom54_mul (mp_ptr pp,
		mp_srcptr ap, mp_size_t an,
		mp_srcptr bp, mp_size_t bn, mp_ptr scratch)
{
  mp_size_t n, s, t;
  int sign;
  TMP_DECL;

  /***************************** decomposition *******************************/
#define a4  (ap + 4 * n)
#define b3  (bp + 3 * n)

  ASSERT (an >= bn);
  n = 1 + (4 * an >= 5 * bn ? (an - 1) / (mp_size_t) 5 : (bn - 1) / (mp_size_t) 4);

  s = an - 4 * n;
  t = bn - 3 * n;

  ASSERT (0 < s && s <= n);
  ASSERT (0 < t && t <= n);
  ASSERT ( s + t > n);
  ASSERT ( s + t > 4);
  ASSERT ( n > 2);

  TMP_MARK;
#define   r8    pp				/* 2n   */
#define   r7    scratch				/* 3n+1 */
#define   r5    (pp + 3*n)			/* 3n+1 */
#define   v0    (pp + 3*n)			/* n+1 */
#define   v1    (pp + 4*n+1)			/* n+1 */
#define   v2    (pp + 5*n+2)			/* n+1 */
#define   v3    (pp + 6*n+3)			/* n+1 */
#define   r3    (scratch + 3 * n + 1)		/* 3n+1 */
#define   r1    (pp + 7*n)			/* s+t <= 2*n */
#define   ws    (scratch + 6 * n + 2)		/* ??? */

  /* Alloc also 3n+1 limbs for ws... mpn_toom_interpolate_8pts may
     need all of them, when DO_mpn_sublsh_n usea a scratch  */
  if(!scratch) scratch = TMP_SALLOC_LIMBS (9 * n + 3);

  /********************** evaluation and recursive calls *********************/
  /* $\pm2$ */
  sign = mpn_toom_ev_lsh (v2, v0, ap, 4, n, s, 1, pp, ws) ^
	 mpn_toom_ev_lsh (v3, v1, bp, 3, n, t, 1, pp, ws);
  TOOM_54_MUL_N_REC(pp, v0, v1, n + 1, ws); /* A(-2)*B(-2) */
  TOOM_54_MUL_N_REC(r3, v2, v3, n + 1, ws+1); /* A(+2)*B(+2) */
  toom_couple_handling(r3, 2*n+1, pp, sign, n, 1, 2);

  /* $\pm1/2$ */
  sign = mpn_toom_ev_rsh (v2, v0, ap, 4, n, s, 1, pp, ws) ^
	 mpn_toom_ev_rsh (v3, v1, bp, 3, n, t, 1, pp, ws);
  TOOM_54_MUL_N_REC(pp, v0, v1, n + 1, ws); /* A(-1/2)*B(-1/2) */
  TOOM_54_MUL_N_REC(r7, v2, v3, n + 1, ws); /* A(+1/2)*B(+1/2) */
  toom_couple_handling(r7, 2*n+1, pp, sign, n, 2, 1);

  /* $\pm1$ */
  sign = mpn_toom_ev_pm1 (v2, v0, ap, 4, n, s,    pp) ^
	 mpn_toom_ev_pm1 (v3, v1, bp, 3, n, t,    pp);
  TOOM_54_MUL_N_REC(pp, v0, v1, n + 1, ws); /* A(-1)*B(-1) */
  TOOM_54_MUL_N_REC(r5, v2, v3, n + 1, ws); /* A(1)*B(1) */
  toom_couple_handling(r5, 2*n+1, pp, sign, n, 0, 0);

  /* A(0)*B(0) */
  TOOM_54_MUL_N_REC(pp, ap, bp, n, ws);

  /* Infinity */
  if(s>t) {
    TOOM_54_MUL_REC(r1, a4, s, b3, t, ws);
  } else {
    TOOM_54_MUL_REC(r1, b3, t, a4, s, ws);
  };

  mpn_toom_interpolate_8pts (pp, n, r3, r7, s + t, ws);

#undef a4
#undef b3
#undef r1
#undef r3
#undef r5
#undef v0
#undef v1
#undef v2
#undef v3
#undef r7
#undef r8
#undef ws
  TMP_FREE;
}

/* Toom-4.5, the splitting 6x3 unbalanced version.
   Evaluate in: infinity, +4, -4, +2, -2, +1, -1, 0.

  <--s-><--n--><--n--><--n--><--n--><--n-->
   ____ ______ ______ ______ ______ ______
  |_a5_|__a4__|__a3__|__a2__|__a1__|__a0__|
                        |b2_|__b1__|__b0__|
                        <-t-><--n--><--n-->

*/
#define TOOM_63_MUL_N_REC(p, a, b, n, ws)		\
  do {	mpn_mul_n(p, a, b, n);				\
  } while (0)

#define TOOM_63_MUL_REC(p, a, na, b, nb, ws)		\
  do {	mpn_mul (p, a, na, b, nb);			\
  } while (0)

void
mpn_toom63_mul (mp_ptr pp,
		mp_srcptr ap, mp_size_t an,
		mp_srcptr bp, mp_size_t bn, mp_ptr scratch)
{
  mp_size_t n, s, t;
  mp_limb_t cy;
  int sign;
  TMP_DECL;

  /***************************** decomposition *******************************/
#define a5  (ap + 5 * n)
#define b0  (bp + 0 * n)
#define b1  (bp + 1 * n)
#define b2  (bp + 2 * n)

  ASSERT (an >= bn);
  n = 1 + (3 * an >= 6 * bn ? (an - 1) / (mp_size_t) 6 : (bn - 1) / (mp_size_t) 3);

  s = an - 5 * n;
  t = bn - 2 * n;

  ASSERT (0 < s && s <= n);
  ASSERT (0 < t && t <= n);
  /* WARNING! it assumes s+t>n */
  ASSERT ( s + t > n );
  ASSERT ( s + t > 4);
  /* WARNING! it assumes n>1 */
  ASSERT ( n > 2);

  TMP_MARK;
#define   r8    pp				/* 2n   */
#define   r7    scratch				/* 3n+1 */
#define   r5    (pp + 3*n)			/* 3n+1 */
#define   v0    (pp + 3*n)			/* n+1 */
#define   v1    (pp + 4*n+1)			/* n+1 */
#define   v2    (pp + 5*n+2)			/* n+1 */
#define   v3    (pp + 6*n+3)			/* n+1 */
#define   r3    (scratch + 3 * n + 1)		/* 3n+1 */
#define   r1    (pp + 7*n)			/* s+t <= 2*n */
#define   ws    (scratch + 6 * n + 2)		/* ??? */

  /* Alloc also 3n+1 limbs for ws... mpn_toom_interpolate_8pts may
     need all of them, when DO_mpn_sublsh_n usea a scratch  */
  if(!scratch) scratch = TMP_SALLOC_LIMBS (9 * n + 3);

  /********************** evaluation and recursive calls *********************/

  /* $\pm2$ */
  sign = mpn_toom_ev_lsh (v2, v0, ap, 5, n, s, 1, pp, ws);
  pp[n] = mpn_lshift(pp, b1, n, 1); /* 2b1 */
  v3[t] = mpn_lshift (v3, b2, t, 2);/* 4b2 */
  if ( n == t )
    v3[n]+= mpn_add_n (v3, v3, b0, n); /* 4b2+b0 */
  else
    v3[n] = mpn_add (v3, b0, n, v3, t+1); /* 4b2+b0 */
  sign ^= abs_sub_add_n(v1, v3, pp, n + 1);
  TOOM_63_MUL_N_REC(pp, v0, v1, n + 1, ws); /* A(-2)*B(-2) */
  TOOM_63_MUL_N_REC(r3, v2, v3, n + 1, ws+1); /* A(+2)*B(+2) */
  toom_couple_handling(r3, 2*n+1, pp, sign, n, 1, 2);

  /* $\pm1/2$ */
  sign = mpn_toom_ev_rsh (v2, v0, ap, 5, n, s, 1, pp, ws) ^
	 mpn_toom_ev_rsh (v3, v1, bp, 2, n, t, 1, pp, ws);
  TOOM_54_MUL_N_REC(pp, v0, v1, n + 1, ws); /* A(-1/2)*B(-1/2) */
  TOOM_54_MUL_N_REC(r7, v2, v3, n + 1, ws); /* A(+1/2)*B(+1/2) */
  toom_couple_handling(r7, 2*n+1, pp, sign, n, 2, 1);


  /* $\pm1$ */
  sign = mpn_toom_ev_pm1 (v2, v0, ap, 5, n, s,    pp);
  /* Compute bs1 and bsm1. Code taken from toom33 */
  cy = mpn_add (ws, b0, n, b2, t);
#if HAVE_NATIVE_mpn_add_n_sub_n
  if (cy == 0 && mpn_cmp (ws, b1, n) < 0)
    {
      cy = mpn_add_n_sub_n (v3, v1, b1, ws, n);
      v3[n] = 0;
      v1[n] = 0;
      sign ^= 1;
    }
  else
    {
      mp_limb_t cy2;
      cy2 = mpn_add_n_sub_n (v3, v1, ws, b1, n);
      w3[n] = cy + (cy2 >> 1);
      w1[n] = cy - (cy & 1);
    }
#else
  v3[n] = cy + mpn_add_n (v3, ws, b1, n);
  if (cy == 0 && mpn_cmp (ws, b1, n) < 0)
    {
      mpn_sub_n (v1, b1, ws, n);
      v1[n] = 0;
      sign ^= 1;
    }
  else
    {
      cy -= mpn_sub_n (v1, ws, b1, n);
      v1[n] = cy;
    }
#endif
  TOOM_63_MUL_N_REC(pp, v0, v1, n + 1, ws); /* A(-1)*B(-1) */
  TOOM_63_MUL_N_REC(r5, v2, v3, n + 1, ws); /* A(1)*B(1) */
  toom_couple_handling(r5, 2*n+1, pp, sign, n, 0, 0);

  /* A(0)*B(0) */
  TOOM_63_MUL_N_REC(pp, ap, bp, n, ws);

  /* Infinity */
  if(s>t) {
    TOOM_63_MUL_REC(r1, a5, s, b2, t, ws);
  } else {
    TOOM_63_MUL_REC(r1, b2, t, a5, s, ws);
  };

  mpn_toom_interpolate_8pts (pp, n, r3, r7, s + t, ws);

#undef a5
#undef b0
#undef b1
#undef b2
#undef r1
#undef r3
#undef r5
#undef v0
#undef v1
#undef v2
#undef v3
#undef r7
#undef r8
#undef ws
  TMP_FREE;
}

/* Toom-4.5, the splitting 6x3 unbalanced version.
   Evaluate in: infinity, +4, -4, +2, -2, +1, -1, 0.

  <--s-><--n--><--n--><--n--><--n--><--n--><--n-->
   ____ ______ ______ ______ ______ ______ ______
  |_a6_|__a5__|__a4__|__a3__|__a2__|__a1__|__a0__|
                                      |b2_|__b0__|
                                      <-t-><--n-->

*/

void
mpn_toom72_mul (mp_ptr pp,
		mp_srcptr ap, mp_size_t an,
		mp_srcptr bp, mp_size_t bn, mp_ptr scratch)
{
  mp_size_t n, s, t;
  mp_limb_t cy;
  int sign;
  TMP_DECL;

  /***************************** decomposition *******************************/
#define a6  (ap + 6 * n)
#define b0  (bp + 0 * n)
#define b1  (bp + 1 * n)

  ASSERT (an >= bn);
  n = 1 + (2 * an >= 7 * bn ? (an - 1) / (mp_size_t) 7 : (bn - 1) / (mp_size_t) 2);

  s = an - 6 * n;
  t = bn - 1 * n;

  ASSERT (0 < s && s <= n);
  ASSERT (0 < t && t <= n);
  ASSERT ( s + t > n );
  ASSERT ( s + t > 4);
  ASSERT ( n > 2);

  TMP_MARK;
#define   r8    pp				/* 2n   */
#define   r7    scratch				/* 3n+1 */
#define   r5    (pp + 3*n)			/* 3n+1 */
#define   v0    (pp + 3*n)			/* n+1 */
#define   v1    (pp + 4*n+1)			/* n+1 */
#define   v2    (pp + 5*n+2)			/* n+1 */
#define   v3    (pp + 6*n+3)			/* n+1 */
#define   r3    (scratch + 3 * n + 1)		/* 3n+1 */
#define   r1    (pp + 7*n)			/* s+t <= 2*n */
#define   ws    (scratch + 6 * n + 2)		/* ??? */

  /* Alloc also 3n+1 limbs for ws... mpn_toom_interpolate_8pts may
     need all of them, when DO_mpn_sublsh_n usea a scratch  */
  if(!scratch) scratch = TMP_SALLOC_LIMBS (9 * n + 3);

  /********************** evaluation and recursive calls *********************/

  /* $\pm2$ */
  sign = mpn_toom_ev_lsh (v2, v0, ap, 6, n, s, 1, pp, ws);
  v1[t] = mpn_lshift(v1, b1, t, 1);
  if(n != t) MPN_ZERO(v1+t+1, n-t);
  v3[n] = v1[n] + mpn_add_n (v3, v1, b0, n);
  if (v1[n] == 0 && mpn_cmp (v1, b0, n) < 0)
    {
      mpn_sub_n (v1, b0, v1, n);
    }
  else
    {
      v1[n] -= mpn_sub_n (v1, v1, b0, n);
      sign ^= 1;
    }
  TOOM_63_MUL_N_REC(pp, v0, v1, n + 1, ws); /* A(-2)*B(-2) */
  TOOM_63_MUL_N_REC(r3, v2, v3, n + 1, ws+1); /* A(+2)*B(+2) */
  toom_couple_handling(r3, 2*n+1, pp, sign, n, 1, 2);

  /* $\pm1/2$ */
  sign = mpn_toom_ev_rsh (v2, v0, ap, 6, n, s, 1, pp, ws);
  v1[n] = mpn_lshift(v1, b0, n, 1);
  mpn_add (v3, v1, n+1, b1, t);
  if ( mpn_zero_p (v1 + t, n + 1 - t) && mpn_cmp (v1, b1, t) < 0)
    {
      mpn_sub_n (v1, b1, v1, t);
      sign ^= 1;
    }
  else
    {
      mpn_sub (v1, v1, n+1, b1, t);
    }
  TOOM_54_MUL_N_REC(pp, v0, v1, n + 1, ws); /* A(-1/2)*B(-1/2) */
  TOOM_54_MUL_N_REC(r7, v2, v3, n + 1, ws); /* A(+1/2)*B(+1/2) */
  toom_couple_handling(r7, 2*n+1, pp, sign, n, 2, 1);

  /* $\pm1$ */
  sign = mpn_toom_ev_pm1 (v2, v0, ap, 6, n, s,    pp);
  /* Compute bs1 and bsm1. Code taken from toom62 */
  if (t == n)
    {
#if HAVE_NATIVE_mpn_add_n_sub_n
      if (mpn_cmp (b0, b1, n) < 0)
	{
	  cy = mpn_add_n_sub_n (v3, v1, b1, b0, n);
	  sign ^= 1;
	}
      else
	{
	  cy = mpn_add_n_sub_n (v3, v1, b0, b1, n);
	}
      v3[n] = cy >> 1;
#else
      v3[n] = mpn_add_n (v3, b0, b1, n);

      if (mpn_cmp (b0, b1, n) < 0)
	{
	  mpn_sub_n (v1, b1, b0, n);
	  sign ^= 1;
	}
      else
	{
	  mpn_sub_n (v1, b0, b1, n);
	}
#endif
    }
  else
    {
      v3[n] = mpn_add (v3, b0, n, b1, t);

      if (mpn_zero_p (b0 + t, n - t) && mpn_cmp (b0, b1, t) < 0)
	{
	  mpn_sub_n (v1, b1, b0, t);
	  MPN_ZERO (v1 + t, n - t);
	  sign ^= 1;
	}
      else
	{
	  mpn_sub (v1, b0, n, b1, t);
	}
    }
  v1[n]=0;

  TOOM_63_MUL_N_REC(pp, v0, v1, n + 1, ws); /* A(-1)*B(-1) */
  TOOM_63_MUL_N_REC(r5, v2, v3, n + 1, ws); /* A(1)*B(1) */
  toom_couple_handling(r5, 2*n+1, pp, sign, n, 0, 0);

  /* A(0)*B(0) */
  TOOM_63_MUL_N_REC(pp, ap, bp, n, ws);

  /* Infinity */
  if(s>t) {
    TOOM_63_MUL_REC(r1, a6, s, b1, t, ws);
  } else {
    TOOM_63_MUL_REC(r1, b1, t, a6, s, ws);
  };

  mpn_toom_interpolate_8pts (pp, n, r3, r7, s + t, ws);

#undef a6
#undef b1
#undef b0
#undef r1
#undef r3
#undef r5
#undef v0
#undef v1
#undef v2
#undef v3
#undef r7
#undef r8
#undef ws
  TMP_FREE;
}

#undef TOOM_54_MUL_N_REC
#undef TOOM_54_MUL_REC

/* Include a default main for checking and timing tests. */
#if defined(CHECK)|defined(TIMING)
#define CONCAT(prefix,M,N,suffix)  prefix ## M ## N ## suffix

#ifdef UNBALANCED
#define M 7
#define N 2
#define mpn_mul_toomMN CONCAT(mpn_toom,7,2,_mul)
#define MINN 3
#else
#define M 5
#define N 4
#define mpn_mul_toomMN CONCAT(mpn_toom,5,4,_mul)
#define MINN 3
#endif

#include "toom-main.c"
#endif
